/*
 * Driver for Data Translation DT2814
 * David A. Schleef
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>



#define DT2814_SIZE 2

#define DT2814_CSR 0
#define DT2814_DATA 1

/*
 * flags
 */

#define DT2814_FINISH 0x80
#define DT2814_ERR 0x40
#define DT2814_BUSY 0x20
#define DT2814_ENB 0x10
#define DT2814_CHANMASK 0x0f

comedi_devinit dt2814_init;
static int dt2814_ad(comedi_device *dev,comedi_trig *it);
static int dt2814_rem(comedi_device *dev);
static void dt2814_interrupt(int irq,void *dev,struct pt_regs * regs);

typedef struct{
	int busy;
	int ntrig;
	int curadchan;
}dt2814_private;
#define devpriv ((dt2814_private *)dev->private)


#define DT2814_TIMEOUT 100

static int dt2814_softtrig(comedi_device *dev,int chan)
{
	comedi_sample s;
	int i,hi,lo;
	unsigned long flags;

	outb(chan,dev->iobase+DT2814_CSR);
	save_flags(flags);
	cli();
	for(i=0;i<DT2814_TIMEOUT;i++){
		if(inb(dev->iobase+DT2814_CSR)&DT2814_FINISH)
			break;
	}
	hi=inb(dev->iobase+DT2814_DATA);
	lo=inb(dev->iobase+DT2814_DATA);
	restore_flags(flags);

	s.chan=chan;
	s.data=(hi<<4)|(lo>>4);
	report_sample(dev,&s);
	
	return s.data;
}

static int dt2814_ad(comedi_device *dev,comedi_trig *it)
{
	if(devpriv->busy)return -EBUSY;

	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		it->data=dt2814_softtrig(dev,it->chan);
		break;
	case COMEDI_TRIGCLK:
		devpriv->ntrig=it->n;
		outb(it->chan|DT2814_ENB|(it->trigvar<<5),
			dev->iobase+DT2814_CSR);
		break;
	default:
		return -EINVAL;
	}
	
	return 0;
}


int dt2814_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,irqs,irq;
	long flags;

	if(strcmp("dt2814",it->name))
		return 0;

	printk("comedi%d: dt2814: 0x%04x ",dev->minor,it->iobase);
	if(check_region(it->iobase,DT2814_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(it->iobase,DT2814_SIZE,"dt2814");
	dev->iobase=it->iobase;
	dev->iosize=DT2814_SIZE;

	outb(0,dev->iobase+DT2814_CSR);
	udelay(100);
	if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
		printk("reset error (fatal)\n");
		return -EIO;
	}
	i=inb(dev->iobase+DT2814_DATA);
	i=inb(dev->iobase+DT2814_DATA);

	irq=it->irq;
	if(irq<0){
		save_flags(flags);
		sti();
		irqs=probe_irq_on();
	
		outb(0,dev->iobase+DT2814_CSR);

		udelay(100);

		irq=probe_irq_off(irqs);
		restore_flags(flags);
		if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
			printk("error probing irq (bad) \n");
		}

		i=inb(dev->iobase+DT2814_DATA);
		i=inb(dev->iobase+DT2814_DATA);
	}
	dev->irq=0;
	if(irq>0){
		printk("( irq = %d )\n",irq);
		request_irq(irq,dt2814_interrupt,SA_INTERRUPT,"dt2814",dev);
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("(probe returned multiple irqs--bad)\n");
	}
	
	dev->chan=16;

	dev->private=kmalloc(sizeof(dt2814_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(dt2814_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(dt2814_private);
	for(i=0;i<dev->chan;i++){
		dev->chinfo[i].trig=dt2814_ad;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}

	comedi_initbuf(&dev->rdbuf);

	dev->rem=dt2814_rem;

	return 1;
}


static int dt2814_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: dt2814: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}


static void dt2814_interrupt(int irq,void *d,struct pt_regs * regs)
{
        int lo,hi;
        comedi_device *dev=d;
        comedi_sample pt;

        hi=inb(dev->iobase+DT2814_DATA);
        lo=inb(dev->iobase+DT2814_DATA);

        pt.data=(hi<<4)|(lo>>4);
        pt.chan=devpriv->curadchan;
	report_sample(dev,&pt);

	if(!(--devpriv->ntrig)){
		int flags,i;

		outb(0,dev->iobase+DT2814_CSR);
		/* note: turning off timed mode triggers another
			sample. */

		save_flags(flags);
		cli();
		for(i=0;i<DT2814_TIMEOUT;i++){
			if(inb(dev->iobase+DT2814_CSR)&DT2814_FINISH)
				break;
		}
		inb(dev->iobase+DT2814_DATA);
		inb(dev->iobase+DT2814_DATA);
		restore_flags(flags);
	}
}


